/*
    TrueTTF is a module for Fenix programming language
    Copyright (C) 2009  Eugene Popov (popov_e_n@mail.ru)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>
*/

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#endif

#include <SDL_ttf.h>

#include <fxdll.h>
//#include <grlib.h>

#include <stdlib.h>
#include <stdio.h>

#define TT_ENC_ASCII 0
#define TT_ENC_UTF8  1
#define TT_ALIGN_H_CENTER 2
#define TT_ALIGN_H_LEFT 4
#define TT_ALIGN_H_RIGHT 8
#define TT_ALIGN_V_MIDDLE 16
#define TT_ALIGN_V_TOP 32
#define TT_ALIGN_V_BOTTOM 64

#define TT_NEW_MAP 128

#define TRUE_TFF_MAXFONTS 20

static struct TTF_Font *true_ttf_fonts[TRUE_TFF_MAXFONTS];
static int true_ttf_initialized = 0;

static int gr_load_truettf (const char * filename, int size)
{
    int i;
    TTF_Font *tmp;

    if(!TTF_WasInit() && TTF_Init()==-1)
    {
        printf("gr_load_truettf: %s\n", TTF_GetError());
    }
    if(!true_ttf_initialized)
    {
        for(i=0; i<TRUE_TFF_MAXFONTS; i++) true_ttf_fonts[i] = NULL;
        true_ttf_initialized = 1;
    }

    tmp = TTF_OpenFont(filename, size);
    if(!tmp) {
        printf("TTF_OpenFont: %s\n", TTF_GetError());
        return -1;
    }

    for(i=0; i<TRUE_TFF_MAXFONTS; i++) {
        if(true_ttf_fonts[i] == NULL) {
            true_ttf_fonts[i] = tmp;
            return i;
        }
    }

    gr_error("gr_load_truettf: No free font slots.\n");
    TTF_CloseFont( tmp );

    return -1;
}

static void gr_unload_truettf (int fontid)
{
    if(fontid>=0 && fontid<TRUE_TFF_MAXFONTS && true_ttf_fonts[fontid]!=NULL)
    {
        TTF_CloseFont( true_ttf_fonts[fontid] );
        true_ttf_fonts[fontid] = NULL;
    }
}

static int gr_write_truettf (GRAPH * dest, int graph, int fontid, int x, int y, int options, const char * text)
{
    int color = 0;
    SDL_Color col = {255,255,255};
    SDL_Surface *text_surface;
    int i,j;

    color = gr_text_getcolor();

    if(fontid>=0 && fontid<TRUE_TFF_MAXFONTS && true_ttf_fonts[fontid]!=NULL)
    {
        if(options & TT_ENC_UTF8)
            text_surface = TTF_RenderUTF8_Solid(true_ttf_fonts[fontid], text, col);
        else
            text_surface = TTF_RenderText_Solid(true_ttf_fonts[fontid], text, col);

        if (text_surface != NULL) {

            int bpp  = text_surface->format->BytesPerPixel;
            int sh = bpp>1 ? 1:0;

            SDL_LockSurface(text_surface);
            Uint8 *cc = (Uint8*) text_surface->pixels;

            if(dest==NULL) {
                dest = bitmap_new_syslib (text_surface->w, text_surface->h, 16, 1) ;
                if (dest) gr_clear (dest);
                if(! dest ) gr_error("gr_write_truettf: cannot allocate NEW_MAP \n");
                graph = dest->code;
            }
            else
            {
                if(options & TT_ALIGN_H_CENTER) x += dest->width/2 - text_surface->w/2;
                if(options & TT_ALIGN_H_RIGHT)  x += dest->width - text_surface->w;
                if(options & TT_ALIGN_V_MIDDLE) y += dest->height/2 - text_surface->h/2;
                if(options & TT_ALIGN_V_BOTTOM) y += dest->height - text_surface->h;
            }

            for(i=0; i<text_surface->h; i++)
            for(j=0; j<text_surface->w; j++)
            {
                int ofs =  bpp*(i*(text_surface->pitch) + j) + sh ;
                if( cc[ofs] ) {
                    gr_put_pixel (dest, x+j, y+i, color) ;
                }
            }
            SDL_UnlockSurface(text_surface);
            SDL_FreeSurface(text_surface);
        }

    }
    else
    {
        gr_error("gr_write_truettf: bad font id\n");
    }

    return graph;
}

static int fxi_load_truettf (INSTANCE * my, int * params)
{
	char * text = (char *)string_get (params[0]) ;
	int r = text ? gr_load_truettf (text, params[1]) : -1 ;
	string_discard (params[0]) ;
	return r ;
}

static int fxi_unload_truettf (INSTANCE * my, int * params)
{
	gr_unload_truettf (params[0]);
	return 0;
}

static int fxi_write_truettf (INSTANCE * my, int * params)
{
    if (!scr_initialized) gr_init (320, 200) ;

    GRAPH * map;
	char * text;
	int graph;

    if(params[5] & TT_NEW_MAP) {
        map = NULL;
        graph = params[0];
    }
    else {
        map = bitmap_get (params[0], params[1]) ;
        graph = params[1];
	}

    text = (char *)string_get (params[6]) ;

    if( text )
        graph = gr_write_truettf (map , graph, params[2], params[3], params[4], params[5], text);

	string_discard (params[6]) ;

	return graph;
}


FENIX_MainDLL RegisterFunctions (COMMON_PARAMS)
{
    FENIX_DLLImport

	FENIX_export ("LOAD_TRUETTF",   "SI",    TYPE_DWORD, fxi_load_truettf );
	FENIX_export ("UNLOAD_TRUETTF", "I",     TYPE_DWORD, fxi_unload_truettf );
	FENIX_export ("WRITE_TRUETTF",  "IIIIIIS",  TYPE_DWORD, fxi_write_truettf );

	if (&fnc_export != 0) {
	    fnc_export ("gr_load_truettf", gr_load_truettf);
	    fnc_export ("gr_unload_truettf", gr_unload_truettf);
	    fnc_export ("gr_write_truettf", gr_write_truettf);
	}
}

